# from os import getenv

# import hub

# from scoach.constants import constants
# from scoach.utils import load_config_file_to_envs

# # TODO: In the future, we could use Hub for data management
# class Dataset (hub.dataset):
#     def __new__(cls, path: str, *args, **kwargs):
#         load_config_file_to_envs()
#         path = f"""s3://{
#             getenv(constants.MINIO_BUCKET_ENV.value)
#         }/{constants.DATASETS_PATH_PREFIX.value}/{path}"""
#         print(path)
#         return super().__new__(
#             cls,
#             path,
#             creds={
#                 "aws_access_key_id": getenv(constants.MINIO_ACCESS_KEY_ENV.value),
#                 "aws_secret_access_key": getenv(constants.MINIO_SECRET_KEY_ENV.value),
#                 "endpoint_url": f"""{getenv(
#                         constants.MINIO_ENDPOINT_SCHEMA_ENV.value
#                     )}://{getenv(
#                         constants.MINIO_ENDPOINT_ENV.value
#                     )}:{getenv(constants.MINIO_ENDPOINT_PORT_ENV.value)}""",
#             },
#             *args,
#             **kwargs
#         )



from os import getenv

from .base import *

# This is not a secret!!!
SECRET_KEY = getenv("SECRET_KEY")

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql_psycopg2",
        "NAME": getenv("DB_NAME"),
        "USER": getenv("DB_USER"),
        "PASSWORD": getenv("DB_PASSWORD"),
        "HOST": getenv("DB_HOST"),
        "PORT": getenv("DB_PORT"),
    }
}



from sys import exit

import minio
import typer
import psycopg2

from scoach.constants import constants
from scoach.utils import load_config_file_to_envs


def check_config():
    """
    Check if the config file exists.
    """
    if not constants.SCOACH_DEFAULT_CONFIG_PATH.value.exists():
        typer.echo(
            "No config file found. Please run `scoach init` to create one.")
        exit(1)

    load_config_file_to_envs()
    return True


def check_database_exists(db_host: str, db_port: str, db_user: str, db_password: str, db_name: str):
    """
    Check if the database exists.
    """
    try:
        conn = psycopg2.connect(
            host=db_host,
            port=db_port,
            database=db_name,
            user=db_user,
            password=db_password
        )
        conn.close()
        return True
    except psycopg2.OperationalError:
        return False


def check_minio_connection(minio_host: str, minio_access_key: str, minio_secret_key: str, minio_bucket: str):
    """
    Check if the minio connection works.
    """
    try:
        minio_client = minio.Minio(
            minio_host,
            access_key=minio_access_key,
            secret_key=minio_secret_key,
            secure=False,
        )
        _ = minio_client.list_objects(minio_bucket)
        return True
    except Exception as e:
        typer.echo(f"Error: {e}")
        return False



"""
Provides scoach, the main class for the application.
"""

import json
import uuid
import time
from os.path import join
from threading import Thread

from scoach.constants import constants
from scoach.executor import Executor
from scoach.logging import logger
from scoach.scheduler import Scheduler
from scoach.utils import get_minio_client, load_env_as_type, save_to_minio


class scoach:
    """
    Main class for scoach
    """

    def __init__(self):
        logger.info("scoach instance initializing...")
        self._scheduler: Scheduler = None
        self._executor: Executor = None
        self._scheduler_thread: Thread = None

    def _initialize_scheduler(self):
        logger.info("Starting Scheduler instance...")
        self._scheduler: Scheduler = Scheduler()
        logger.info("Scheduler instance started successfully!")

    def _initialize_executor(self):
        logger.info("Starting Executor instance...")
        if not self._scheduler:
            self._initialize_scheduler()
        self._executor: Executor = Executor(self._scheduler)
        logger.info("Executor instance started successfully!")

    def _initialize_scheduler_thread(self):
        logger.info("Starting Scheduler thread...")
        self._scheduler_thread = Thread(target=self._threaded_scheduler)
        self._scheduler_thread.start()
        logger.info("Scheduler thread started successfully!")

    @logger.catch
    def _threaded_scheduler(self):
        """
        Constantly checks for new runs
        and launches them
        :return:
        """
        from scoach.models import Run
        while True:
            try:
                new_runs = Run.objects.all().filter(
                    status__status=constants.RUN_STATUS_CREATED.value)
                for run in new_runs:
                    logger.info(f"Sending new run ({run}) to the executor")
                    if not self._scheduler:
                        self._initialize_scheduler()
                    if not self._executor:
                        self._initialize_executor()
                    self._executor.execute(run.id)
            except Exception as e:
                logger.error(f"Error in scheduler: {e}")
            finally:
                logger.info(
                    f"Scheduler sleeping for {constants.SCHEDULER_SLEEP_TIME.value} seconds")
                time.sleep(constants.SCHEDULER_SLEEP_TIME.value)

    @logger.catch
    def start_scheduler(self):
        """
        Starts the scoach daemon
        :return:
        """
        if not self._scheduler:
            self._initialize_scheduler()
        if not self._executor:
            self._initialize_executor()
        logger.info("Starting scheduler in a thread...")
        self._initialize_scheduler_thread()
        logger.info("scoach daemon is executing...")



"""
General utilities
"""

import os
from os.path import join
from os import makedirs
from typing import Any
from sys import exit

import yaml
import typer
from minio import Minio
import django
from django.db import models
from django.core import management

from scoach.constants import constants
from scoach.logging import logger


def load_env_as_type(env_name, env_type=str, default=None):
    """
    Loads an environment variable as a certain type.
    """
    env_value = os.environ.get(env_name)
    # pylint: disable=no-else-return
    if env_value is None:
        return default
    elif env_type == int:
        return int(env_value)
    elif env_type == bool:
        return env_value.lower() in ["true", "yes", "1", "y"]
    elif env_type == str:
        return env_value
    else:
        raise ValueError("Unknown type: %s" % env_type)


def get_minio_client(
    minio_access_key: str = None,
    minio_secret_key: str = None,
    minio_url: str = None,
    minio_bucket: str = None,
):
    """
    Get a Minio client for the given Minio URL and bucket.
    """
    minio_access_key = minio_access_key or load_env_as_type(
        constants.MINIO_ACCESS_KEY_ENV.value,
        str,
        constants.MINIO_ACCESS_KEY_ENV_DEFAULT.value,
    )
    minio_secret_key = minio_secret_key or load_env_as_type(
        constants.MINIO_SECRET_KEY_ENV.value,
        str,
        constants.MINIO_SECRET_KEY_ENV_DEFAULT.value,
    )
    minio_url = minio_url or load_env_as_type(
        constants.MINIO_ENDPOINT_ENV.value,
        str,
        constants.MINIO_ENDPOINT_ENV_DEFAULT.value,
    )
    minio_bucket = minio_bucket or load_env_as_type(
        constants.MINIO_BUCKET_ENV.value, str, constants.MINIO_BUCKET_ENV_DEFAULT.value
    )
    if not minio_url:
        raise ValueError("Missing Minio URL")
    if not minio_bucket:
        raise ValueError("Missing Minio bucket")
    return Minio(minio_url, access_key=minio_access_key, secret_key=minio_secret_key)


def join_tags(tags: list) -> str:
    """
    Merge tags into a single string.
    """
    return ",".join(tags)


def split_tags(tags: str) -> list:
    """
    Split tags into a list of tags
    """
    return tags.split(",")


def load_config_file_to_envs():
    """
    Loads the config file to environment variables.
    """
    if not constants.SCOACH_DEFAULT_CONFIG_PATH.value.exists():
        typer.echo(
            "No config file found. Please run `scoach init` to create one.")
        exit(1)
    # pylint: disable=unspecified-encoding
    # pylint: disable=invalid-name
    with open(constants.SCOACH_DEFAULT_CONFIG_PATH.value, "r") as f:
        config: dict = yaml.safe_load(f)
    for key, value in config.items():
        os.environ[key] = str(value)


def save_to_minio(minio_client: Minio, file_path: str, file_name: str):
    """
    Save a file to MinIO
    """
    minio_client.fput_object(
        load_env_as_type(
            constants.MINIO_BUCKET_ENV.value,
            default=constants.MINIO_BUCKET_ENV_DEFAULT.value,
        ),
        file_path,
        file_name,
    )


def download_from_minio(minio_client: Minio, file_path: str, file_name: str):
    """
    Download a file from MinIO
    """
    minio_client.fget_object(
        load_env_as_type(
            constants.MINIO_BUCKET_ENV.value,
            default=constants.MINIO_BUCKET_ENV_DEFAULT.value,
        ),
        file_path,
        file_name,
    )


def safe_object_get(T: models.Model, **kwargs):
    """
    Safely get an object from the database.
    """
    try:
        return T.objects.get(**kwargs)
    except T.DoesNotExist:
        return None


def safe_cast(value: str, to_type: type, default: Any = None):
    """
    Safely cast a value to a given type.
    """
    try:
        return to_type(value)
    except ValueError:
        return default


def save_run(run_id: str, model, train_score: float, validation_score: float):
    """
    Save a run to the database.
    """
    from scoach.models import Run, Weights
    # Cast types
    train_score = safe_cast(train_score, float, None)
    if train_score is None:
        raise ValueError("Train score is not a float")
    validation_score = safe_cast(validation_score, float, None)
    if validation_score is None:
        raise ValueError("Validation score is not a float")
    # Ensure run exists
    run = safe_object_get(Run, id=run_id)
    if run is None:
        raise ValueError(f"Run #{run_id} does not exist")
    # Generate unique weights_path
    weights_path = join(constants.WEIGHTS_PATH_PREFIX.value, f"{run_id}.h5")
    tmp_weights_file = join("/tmp/", weights_path)
    makedirs(tmp_weights_file.rsplit("/", 1)[0], exist_ok=True)
    # Save weights locally
    model.save_weights(tmp_weights_file)
    # Save weights on MinIO
    minio_client = get_minio_client()
    save_to_minio(minio_client, weights_path, tmp_weights_file)
    # Save run on DB
    run: Run = safe_object_get(Run, id=run_id)
    if run:
        run.train_score = train_score
        run.validation_score = validation_score
        weights: Weights = Weights.objects.create(
            path=weights_path,
        )
        weights.save()
        run.weights = weights
        run.save()
    else:
        raise ValueError("Run does not exist")


def load_run(run_id: str):
    """
    Load a run from the database.
    """
    # pylint: disable=no-name-in-module
    # pylint: disable=import-outside-toplevel
    from tensorflow.keras.models import model_from_json
    from scoach.models import Run, Weights

    run: Run = safe_object_get(Run, id=run_id)
    if run:
        weights: Weights = run.weights
        weights_path = weights.path
        # Download weights from MinIO
        minio_client = get_minio_client()
        tmp_weights_file = join("/tmp/", weights_path)
        download_from_minio(minio_client, weights_path,
                            tmp_weights_file)
        # Load weights locally
        model = model_from_json(run.model.config)
        model.load_weights(tmp_weights_file)
        # Remove weights locally
        os.remove(join("/tmp/", weights_path))
        return model
    else:
        raise ValueError("Run not found")


def setup_django():
    load_config_file_to_envs()
    mode = os.getenv(constants.DJANGO_SETTINGS_MODE_ENV.value,
                     constants.DJANGO_SETTINGS_MODE_ENV_DEFAULT.value)
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', f'scoach.settings.{mode}')
    try:
        django.setup()
    except RuntimeError as e:
        if "populate() isn't reentrant" not in str(e):
            raise e


def setup_database():
    setup_django()
    management.execute_from_command_line(
        ['manage.py', 'makemigrations', 'scoach'])
    management.execute_from_command_line(['manage.py', 'migrate'])



"""
Provides CLI app for scoach.
"""

import json
from typing import List
import uuid
from pathlib import Path
from os.path import join

import typer

from scoach.cli.utils import check_config
from scoach.constants import constants
from scoach.logging import logger
from scoach.utils import (
    get_minio_client,
    save_to_minio,
    safe_object_get
)

app = typer.Typer()


@app.command()
@logger.catch
def submit(python_script: Path, job_config: Path, model_config: Path, tags: List[str] = None):
    """
    Submits a job to the queue.
    """
    if not check_config():
        return
    from scoach.models import Script, Parameters, Model, Status, Run, Tag
    # Check if all paths exists
    if not python_script.exists():
        script: Script = safe_object_get(Script, id=str(python_script))
        if script is None:
            typer.echo(
                "Python script does not exist locally and was not found on database."
            )
            return
    else:
        script_id = str(uuid.uuid4())
        minio_script_path = join(
            constants.SCRIPTS_PATH_PREFIX.value,
            script_id,
            str(python_script).split("/")[-1],
        )
        script: Script = Script.objects.create(
            path=minio_script_path
        )
        save_to_minio(get_minio_client(),
                      minio_script_path, str(python_script))
        script.save()
    if not job_config.exists():
        typer.echo("Job config does not exist.")
        return
    if not model_config.exists():
        typer.echo("Model config does not exist.")
        return
    # Load job_config json
    # pylint: disable=invalid-name
    with job_config.open() as f:
        job_config = json.load(f)
    parameters: Parameters = safe_object_get(
        Parameters, config=json.dumps(job_config))
    if parameters is None:
        parameters = Parameters.objects.create(config=json.dumps(job_config))
        parameters.save()
    # Load model_config json
    # pylint: disable=invalid-name
    with model_config.open() as f:
        model_config = json.load(f)
    model: Model = safe_object_get(Model, config=json.dumps(model_config))
    if model is None:
        model = Model.objects.create(config=json.dumps(model_config))
    model.save()
    # Submit job
    status: Status = safe_object_get(
        Status, status=constants.RUN_STATUS_CREATED.value)
    if status is None:
        status = Status.objects.create(
            status=constants.RUN_STATUS_CREATED.value)
        status.save()
    run: Run = Run.objects.create(
        model=model,
        script=script,
        parameters=parameters,
        status=status,
    )
    run.save()
    if tags is not None:
        for tag in tags:
            tag_obj: Tag = safe_object_get(Tag, name=tag)
            if tag_obj is None:
                tag_obj: Tag = Tag.objects.create(name=tag)
                tag_obj.save()
            run.tags.add(tag_obj)
        run.save()
    typer.echo(f"Submitted run with id {run.id}")


# pylint: disable=redefined-builtin
@app.command()
@logger.catch
def list():
    """
    List all jobs.
    """
    if not check_config():
        return
    from scoach.models import Run
    runs = Run.objects.all()
    typer.echo(f"Found {len(runs)} runs.")
    for run in runs:
        typer.echo(f"  - {run}")


@app.command()
@logger.catch
def delete(run_id: str):
    """
    Deletes a run from the database.
    """
    if not check_config():
        return
    from scoach.models import Run
    run: Run = safe_object_get(Run, id=run_id)
    if run is None:
        typer.echo(f"Run {run_id} not found.")
        return
    typer.echo(f"Run {run_id} will be deleted!")
    typer.echo(f"Information: {run}")
    choice = typer.prompt("Are you sure?", default=False,
                          show_choices=True, type=bool)
    if choice:
        run.delete()
        typer.echo(f"Run {run_id} deleted!")
    else:
        typer.echo("Run not deleted.")


@app.command()
@logger.catch
def describe(run_id: str):
    """
    Describes a run from the database.
    """
    if not check_config():
        return
    from scoach.models import Run
    run: Run = safe_object_get(Run, id=run_id)
    if run is None:
        typer.echo(f"Run {run_id} not found.")
        return
    typer.echo(f"Run #{run_id}")
    typer.echo(f"  - Status: {run.status}")
    typer.echo(f"  - Train score: {run.train_score}")
    typer.echo(f"  - Validation score: {run.validation_score}")
    typer.echo(f"  - Created at: {run.date_created}")
    typer.echo(f"  - Modified at: {run.date_modified}")
    typer.echo(f"  - Model: {run.model}")
    typer.echo(f"  - Script: {run.script}")
    typer.echo(f"  - Parameters: {run.parameters}")
    typer.echo(f"  - Tags: {[t.name for t in run.tags.all()]}")
    typer.echo()
    typer.echo(
        "If you desire to load this model, open a Python shell and do the following:"
    )
    typer.echo(">>> from scoach.utils import load_run")
    typer.echo(f">>> model = load_run({run_id})")


@app.command()
@logger.catch
def retry(run_id: str):
    """
    Retries a run.
    """
    if not check_config():
        return
    from scoach.models import Status, Run
    run: Run = safe_object_get(Run, id=run_id)
    if run is None:
        typer.echo(f"Run {run_id} not found.")
        return
    if run.status.status != constants.RUN_STATUS_FAILED.value:
        typer.echo(f"Run {run_id} is not in failed status.")
        return
    status: Status = safe_object_get(
        Status, status=constants.RUN_STATUS_CREATED.value)
    if status is None:
        status = Status.objects.create(
            status=constants.RUN_STATUS_CREATED.value)
        status.save()
    run.status = status
    run.save()
    typer.echo(f"Run {run_id} retried!")



"""
Provides constants for the application.
"""

from enum import Enum
from os import getenv
from pathlib import Path


class constants(Enum):  # pylint: disable=invalid-name
    """
    Constants for the application
    """

    # Time
    SECOND = 1
    MINUTE = 60 * SECOND
    HOUR = 60 * MINUTE
    DAY = 24 * HOUR
    WEEK = 7 * DAY
    MONTH = 30 * DAY
    YEAR = 365 * DAY

    # Redis
    REDIS_HOST_ENV = "REDIS_HOST"
    REDIS_HOST_ENV_DEFAULT = "localhost"
    REDIS_PORT_ENV = "REDIS_PORT"
    REDIS_PORT_ENV_DEFAULT = 6379
    REDIS_DB_ENV = "REDIS_DB"
    REDIS_DB_ENV_DEFAULT = 0
    REDIS_PASSWORD_ENV = "REDIS_PASSWORD"
    REDIS_PASSWORD_ENV_DEFAULT = None
    REDIS_QUEUE_NAME_ENV = "REDIS_QUEUE_NAME"
    REDIS_QUEUE_NAME_ENV_DEFAULT = None

    # Database
    DB_HOST_ENV = "DB_HOST"
    DB_HOST_ENV_DEFAULT = None
    DB_PORT_ENV = "DB_PORT"
    DB_PORT_ENV_DEFAULT = 5432
    DB_USER_ENV = "DB_USER"
    DB_USER_ENV_DEFAULT = "postgres"
    DB_PASSWORD_ENV = "DB_PASSWORD"
    DB_PASSWORD_ENV_DEFAULT = None
    DB_NAME_ENV = "DB_NAME"
    DB_NAME_ENV_DEFAULT = "scoach"
    RUN_STATUS_CREATED = "CREATED"
    RUN_STATUS_QUEUED = "QUEUED"
    RUN_STATUS_RUNNING = "RUNNING"
    RUN_STATUS_COMPLETED = "COMPLETED"
    RUN_STATUS_FAILED = "FAILED"
    RUN_STATUS_CANCELED = "CANCELED"

    # SLURM
    SLURM_PARTITION_ENV = "SLURM_PARTITION"
    SLURM_PARTITION_ENV_DEFAULT = None
    SLURM_CORES_PER_JOB_ENV = "SLURM_CORES_PER_JOB"
    SLURM_CORES_PER_JOB_ENV_DEFAULT = 1
    SLURM_MEMORY_PER_JOB_ENV = "SLURM_MEMORY_PER_JOB"
    SLURM_MEMORY_PER_JOB_ENV_DEFAULT = "1G"
    SLURM_WORKER_NAME_ENV = "SLURM_WORKER_NAME"
    SLURM_WORKER_NAME_ENV_DEFAULT = None
    SLURM_JOB_EXCLUSIVE_ENV = "SLURM_JOB_EXCLUSIVE"
    SLURM_JOB_EXCLUSIVE_ENV_DEFAULT = False
    SLURM_MAX_WORKERS_ENV = "SLURM_MAX_WORKERS"
    SLURM_MAX_WORKERS_ENV_DEFAULT = 1

    # MinIO
    MINIO_ENDPOINT_ENV = "MINIO_ENDPOINT"
    MINIO_ENDPOINT_ENV_DEFAULT = None
    MINIO_ENDPOINT_SCHEMA_ENV = "MINIO_ENDPOINT_SCHEMA"
    MINIO_ENDPOINT_SCHEMA_ENV_DEFAULT = "https"
    MINIO_ENDPOINT_PORT_ENV = "MINIO_ENDPOINT_PORT"
    MINIO_ENDPOINT_PORT_ENV_DEFAULT = 443
    MINIO_ACCESS_KEY_ENV = "MINIO_ACCESS_KEY"
    MINIO_ACCESS_KEY_ENV_DEFAULT = None
    MINIO_SECRET_KEY_ENV = "MINIO_SECRET_KEY"
    MINIO_SECRET_KEY_ENV_DEFAULT = None
    MINIO_BUCKET_ENV = "MINIO_BUCKET"
    MINIO_BUCKET_ENV_DEFAULT = None
    WEIGHTS_PATH_PREFIX = "weights/"
    SCRIPTS_PATH_PREFIX = "scripts/"
    DATASETS_PATH_PREFIX = "datasets"

    # Dask scheduler
    DASK_DEFAULT_WORKERS = 0

    # scoach
    SCOACH_DEFAULT_CONFIG_PATH = Path.home() / ".scoach/config.yaml"
    SCHEDULER_SLEEP_TIME = 30 * SECOND

    # Django
    DJANGO_SETTINGS_MODE_ENV = "DJANGO_SETTINGS_MODE"
    DJANGO_SETTINGS_MODE_ENV_DEFAULT = "prod"



import os

import django
from django.db import models
from django.core import management

from scoach.constants import constants
from scoach.utils import load_config_file_to_envs, setup_django


setup_django()


class Model(models.Model):
    id = models.AutoField(primary_key=True)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    description = models.TextField(blank=True, null=True)
    config = models.TextField(blank=True, null=True)

    def __str__(self):
        return f"Model #{self.id} (desc={self.description},created={self.date_created})"


class Script(models.Model):
    id = models.AutoField(primary_key=True)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    description = models.TextField(blank=True, null=True)
    path = models.TextField()

    def __str__(self):
        return f"Script #{self.id} (desc={self.description},created={self.date_created})"


class Parameters(models.Model):
    id = models.AutoField(primary_key=True)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    description = models.TextField(blank=True, null=True)
    config = models.TextField(blank=True, null=True)

    def __str__(self):
        return f"Parameters #{self.id} (desc={self.description},created={self.date_created})"


class Weights(models.Model):
    id = models.AutoField(primary_key=True)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    description = models.TextField(blank=True, null=True)
    path = models.TextField()

    def __str__(self):
        return f"Weights #{self.id} (desc={self.description},created={self.date_created})"


class Status(models.Model):
    id = models.AutoField(primary_key=True)
    status = models.TextField()
    description = models.TextField(blank=True, null=True)

    def __str__(self):
        return f"{self.status}"


class Tag(models.Model):
    id = models.AutoField(primary_key=True)
    name = models.TextField()
    description = models.TextField(blank=True, null=True)

    def __str__(self):
        return f"Tag {self.name} (desc={self.description})"


class Run(models.Model):
    id = models.AutoField(primary_key=True)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    description = models.TextField(blank=True, null=True)
    model = models.ForeignKey(Model, on_delete=models.CASCADE)
    script = models.ForeignKey(Script, on_delete=models.CASCADE)
    parameters = models.ForeignKey(Parameters, on_delete=models.CASCADE)
    weights = models.ForeignKey(Weights, on_delete=models.CASCADE, null=True)
    status = models.ForeignKey(Status, on_delete=models.CASCADE)
    tags = models.ManyToManyField(Tag)
    train_score = models.FloatField(blank=True, null=True)
    validation_score = models.FloatField(blank=True, null=True)

    def __str__(self):
        return f"Run #{self.id} (status={self.status},tags={[t.name for t in self.tags.all()]},created={self.date_created})"


# class Dataset(models.Model):
#     id = models.AutoField(primary_key=True)
#     date_created = models.DateTimeField(auto_now_add=True)
#     date_modified = models.DateTimeField(auto_now=True)
#     description = models.TextField(blank=True, null=True)
#     path = models.TextField()

#     def __str__(self):
#         return f"Dataset #{self.id} (desc={self.description},path={self.path},created={self.date_created})"



from pathlib import Path

import yaml
import typer

from scoach.cli import run, script
from scoach.cli.utils import (
    check_config,
    check_database_exists,
    check_minio_connection,
)
from scoach.scoach import scoach
from scoach.constants import constants
from scoach.logging import logger
from scoach.utils import setup_database

app = typer.Typer()
app.add_typer(run.app, name="run", help="Manages runs")
app.add_typer(script.app, name="script", help="Manages scripts")


@app.command()
@logger.catch
def init():  # pylint: disable=too-many-locals
    """
    Initialize (or override) configuration file.
    """
    # DB_HOST: "postgres"
    # DB_PORT: 5432
    # DB_USER: "postgres"
    # DB_PASSWORD: "postgres"
    # DB_NAME: "scoach"
    # SLURM_PARTITION: "debug"
    # SLURM_CORES_PER_JOB: 1
    # SLURM_MEMORY_PER_JOB: "1G"
    # SLURM_WORKER_NAME: "worker_name"
    # SLURM_JOB_EXCLUSIVE: false
    # SLURM_MAX_WORKERS: 1
    # MINIO_ENDPOINT: "minio.example.com"
    # MINIO_ENDPOINT_SCHEMA: "https"
    # MINIO_ENDPOINT_PORT: 443
    # MINIO_ACCESS_KEY: "minio_access_key"
    # MINIO_SECRET_KEY: "minio_secret_key"
    # MINIO_BUCKET: "minio_bucket_name"
    typer.echo(
        "Welcome to scoach! Please fill all requested fields!")
    db_host = typer.prompt(constants.DB_HOST_ENV.value,
                           default=constants.DB_HOST_ENV_DEFAULT.value)
    db_port = typer.prompt(constants.DB_PORT_ENV.value,
                           default=constants.DB_PORT_ENV_DEFAULT.value)
    db_user = typer.prompt(constants.DB_USER_ENV.value,
                           default=constants.DB_USER_ENV_DEFAULT.value)
    db_password = typer.prompt(
        constants.DB_PASSWORD_ENV.value, default=constants.DB_PASSWORD_ENV_DEFAULT.value)
    db_name = typer.prompt(constants.DB_NAME_ENV.value,
                           default=constants.DB_NAME_ENV_DEFAULT.value)

    typer.echo("Checking database connection...")
    if check_database_exists(db_host, db_port, db_user, db_password, db_name):
        typer.echo("Database connection successful!")
    else:
        typer.echo(
            "Database connection failed, please check your inputs or create database if needed.")
        return

    slurm_partition = typer.prompt(
        constants.SLURM_PARTITION_ENV.value, default=constants.SLURM_PARTITION_ENV_DEFAULT.value)
    slurm_cores_per_job = typer.prompt(
        constants.SLURM_CORES_PER_JOB_ENV.value, default=constants.SLURM_CORES_PER_JOB_ENV_DEFAULT.value)
    slurm_memory_per_job = typer.prompt(
        constants.SLURM_MEMORY_PER_JOB_ENV.value, default=constants.SLURM_MEMORY_PER_JOB_ENV_DEFAULT.value)
    slurm_worker_name = typer.prompt(
        constants.SLURM_WORKER_NAME_ENV.value, default=constants.SLURM_WORKER_NAME_ENV_DEFAULT.value)
    slurm_job_exclusive = typer.confirm(
        constants.SLURM_JOB_EXCLUSIVE_ENV.value, default=constants.SLURM_JOB_EXCLUSIVE_ENV_DEFAULT.value)
    slurm_max_workers = typer.prompt(
        constants.SLURM_MAX_WORKERS_ENV.value, default=constants.SLURM_MAX_WORKERS_ENV_DEFAULT.value)

    minio_endpoint = typer.prompt(
        constants.MINIO_ENDPOINT_ENV.value, default=constants.MINIO_ENDPOINT_ENV_DEFAULT.value)
    minio_endpoint_schema = typer.prompt(
        constants.MINIO_ENDPOINT_SCHEMA_ENV.value, default=constants.MINIO_ENDPOINT_SCHEMA_ENV_DEFAULT.value)
    minio_endpoint_port = typer.prompt(
        constants.MINIO_ENDPOINT_PORT_ENV.value, default=constants.MINIO_ENDPOINT_PORT_ENV_DEFAULT.value)
    minio_access_key = typer.prompt(
        constants.MINIO_ACCESS_KEY_ENV.value, default=constants.MINIO_ACCESS_KEY_ENV_DEFAULT.value)
    minio_secret_key = typer.prompt(
        constants.MINIO_SECRET_KEY_ENV.value, default=constants.MINIO_SECRET_KEY_ENV_DEFAULT.value)
    minio_bucket = typer.prompt(
        constants.MINIO_BUCKET_ENV.value, default=constants.MINIO_BUCKET_ENV_DEFAULT.value)

    typer.echo("Checking MinIO connection...")
    if check_minio_connection(minio_endpoint, minio_access_key, minio_secret_key, minio_bucket):
        typer.echo("MinIO connection successful!")
    else:
        typer.echo(
            "MinIO connection failed, please check your inputs or create bucket if needed.")
        return

    # Save configs to YAML file
    config_path = Path(constants.SCOACH_DEFAULT_CONFIG_PATH.value)
    config_path.parent.mkdir(parents=True, exist_ok=True)
    config_path.write_text(
        yaml.dump(
            {
                constants.DB_HOST_ENV.value: db_host,
                constants.DB_PORT_ENV.value: db_port,
                constants.DB_USER_ENV.value: db_user,
                constants.DB_PASSWORD_ENV.value: db_password,
                constants.DB_NAME_ENV.value: db_name,
                constants.SLURM_PARTITION_ENV.value: slurm_partition,
                constants.SLURM_CORES_PER_JOB_ENV.value: slurm_cores_per_job,
                constants.SLURM_MEMORY_PER_JOB_ENV.value: slurm_memory_per_job,
                constants.SLURM_WORKER_NAME_ENV.value: slurm_worker_name,
                constants.SLURM_JOB_EXCLUSIVE_ENV.value: slurm_job_exclusive,
                constants.SLURM_MAX_WORKERS_ENV.value: slurm_max_workers,
                constants.MINIO_ENDPOINT_ENV.value: minio_endpoint,
                constants.MINIO_ENDPOINT_SCHEMA_ENV.value: minio_endpoint_schema,
                constants.MINIO_ENDPOINT_PORT_ENV.value: minio_endpoint_port,
                constants.MINIO_ACCESS_KEY_ENV.value: minio_access_key,
                constants.MINIO_SECRET_KEY_ENV.value: minio_secret_key,
                constants.MINIO_BUCKET_ENV.value: minio_bucket,
                constants.DJANGO_SETTINGS_MODE_ENV.value: constants.DJANGO_SETTINGS_MODE_ENV_DEFAULT.value,
            },
        )
    )
    typer.echo(
        f"Config file created at {constants.SCOACH_DEFAULT_CONFIG_PATH.value}.")

    typer.echo("Setting up database...")
    setup_database()


@app.command()
@logger.catch
def start():
    """
    Run daemon process for SLURM login machine.
    """
    if not check_config():
        return

    typer.echo("Starting scoach daemon process...")
    scoach = scoach()
    scoach.start_scheduler()



"""
Provides Executor, a wrapper for a Prefect Flow
"""

import json
from time import sleep
from threading import Thread
from functools import partial

import jinja2
import prefect
from prefect import task, Flow
from prefect.engine.state import State
from prefect.executors import DaskExecutor

from scoach.constants import constants
from scoach.logging import logger
from scoach.scheduler import Scheduler
from scoach.utils import get_minio_client, load_config_file_to_envs, load_env_as_type, safe_object_get


@task
def setup():
    load_config_file_to_envs()


@task
def update_run_status(run_id: str):
    """
    Task to update the status of a run

    Args:
        - run (Run): the run to update
    """
    from scoach.models import Status, Run
    status = safe_object_get(
        Status, status=constants.RUN_STATUS_RUNNING.value)
    if status is None:
        status = Status(status=constants.RUN_STATUS_RUNNING.value)
        status.save()
    run: Run = safe_object_get(Run, id=run_id)
    run.status = status
    run.save()


@task
def render_template(template_text: str, **kwargs):
    """
    Task to render a jinja2 template

    Args:
        - template_text (str): the text of the template to render
        - **kwargs (dict): the keyword arguments to pass to the template

    Returns:
        - str: the rendered template
    """
    return jinja2.Template(template_text).render(**kwargs)


@task
def execute_template(rendered_template: str):
    """
    Task to execute a rendered template

    Args:
        - rendered_template (str): the rendered template to execute
    """
    logger = prefect.context.get("logger")
    logger.info(f"Executing template: {rendered_template}")
    # pylint: disable=exec-used
    exec(rendered_template)


class Executor:  # pylint: disable=too-few-public-methods
    """
    A wrapper for a Prefect Flow
    """

    def __init__(self, scheduler: Scheduler):
        self._scheduler = scheduler
        self._executor = DaskExecutor(scheduler.address)

    def _threaded_execution_handler(self, flow: Flow, run_id: str):
        """
        Threaded handler for the flow execution
        """
        from scoach.models import Status, Run
        status = safe_object_get(
            Status, status=constants.RUN_STATUS_QUEUED.value)
        if status is None:
            status = Status(status=constants.RUN_STATUS_QUEUED.value)
            status.save()
        run: Run = safe_object_get(Run, id=run_id)
        run.status = status
        run.save()
        state: State = flow.run()  # (executor=self._executor)
        while not state.is_finished():
            sleep(0.25)
        if state.is_failed():
            status = safe_object_get(
                Status, status=constants.RUN_STATUS_FAILED.value)
            if status is None:
                status = Status(
                    status=constants.RUN_STATUS_FAILED.value)
                status.save()
            run: Run = safe_object_get(Run, id=run_id)
            run.status = status
            run.save()
            logger.error(
                f"Job {run_id} failed with error: {state.result}")
        elif state.is_successful():
            status = safe_object_get(
                Status, status=constants.RUN_STATUS_COMPLETED.value)
            if status is None:
                status = Status(
                    status=constants.RUN_STATUS_COMPLETED.value)
                status.save()
            run: Run = safe_object_get(Run, id=run_id)
            run.status = status
            run.save()
            logger.info(f"Job {run_id} completed successfully")

    def execute(self, run_id: str):
        """
        Execute the training job
        """
        from scoach.models import Model, Run
        with Flow("Training Flow") as flow:
            setup()
            update_run_status(run_id=run_id)
            minio_client = get_minio_client()
            bucket = load_env_as_type(
                constants.MINIO_BUCKET_ENV.value,
                default=constants.MINIO_BUCKET_ENV_DEFAULT.value,
            )
            run: Run = safe_object_get(Run, id=run_id)
            script_path = run.script.path
            template_text = (
                minio_client.get_object(
                    bucket, script_path).read().decode("utf-8")
            )
            model: Model = run.model
            logger.info(model.config)
            rendered_template = render_template(
                template_text,
                run_id=run_id,
                model_config=f"\"\"\"{model.config}\"\"\"",
                **json.loads(run.parameters.config),
            )
            execute_template(rendered_template)

        thread = Thread(
            target=partial(self._threaded_execution_handler, flow, run_id),
            daemon=True,
        )
        thread.start()



"""
Provides the Scheduler class
"""

from dask_jobqueue import SLURMCluster

from scoach.constants import constants
from scoach.utils import load_env_as_type


class Scheduler:  # pylint: disable=too-few-public-methods
    """
    Scheduler class

    Essentially a wrapper for the Dask
    SLURMCluster class.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        partition=None,
        cores=None,
        memory=None,
        name=None,
        exclusive=None,
        max_workers=None,
    ):
        partition = partition or load_env_as_type(
            constants.SLURM_PARTITION_ENV.value,
            default=constants.SLURM_PARTITION_ENV_DEFAULT.value,
        )
        cores = cores or load_env_as_type(
            constants.SLURM_CORES_PER_JOB_ENV.value,
            int,
            default=constants.SLURM_CORES_PER_JOB_ENV_DEFAULT.value,
        )
        memory = memory or load_env_as_type(
            constants.SLURM_MEMORY_PER_JOB_ENV.value,
            default=constants.SLURM_MEMORY_PER_JOB_ENV_DEFAULT.value,
        )
        name = name or load_env_as_type(
            constants.SLURM_WORKER_NAME_ENV.value,
            default=constants.SLURM_WORKER_NAME_ENV_DEFAULT.value,
        )
        exclusive = exclusive or load_env_as_type(
            constants.SLURM_JOB_EXCLUSIVE_ENV.value,
            bool,
            default=constants.SLURM_JOB_EXCLUSIVE_ENV_DEFAULT.value,
        )
        max_workers = max_workers or load_env_as_type(
            constants.SLURM_MAX_WORKERS_ENV.value,
            int,
            default=constants.SLURM_MAX_WORKERS_ENV_DEFAULT.value,
        )
        job_extra = ["--exclusive"] if exclusive else []
        self._cluster = SLURMCluster(
            queue=partition, cores=cores, memory=memory, name=name, job_extra=job_extra,
        )
        self._cluster.scale(jobs=constants.DASK_DEFAULT_WORKERS.value)
        self._cluster.adapt(maximum_jobs=max_workers)

    @property
    def address(self):
        """
        Address for the cluster scheduler
        """
        return self._cluster.scheduler.address




import typer

from scoach.cli.utils import check_config
from scoach.logging import logger
from scoach.utils import (
    safe_object_get,
    get_minio_client,
    download_from_minio,
)

app = typer.Typer()


@app.command()
@logger.catch
def list():
    """
    List all scripts in the database.
    """
    if not check_config():
        return
    from scoach.models import Script
    scripts = Script.objects.all()
    if len(scripts) == 0:
        typer.echo("No scripts found.")
        return
    typer.echo(f"Found {len(scripts)} scripts.")
    for script in scripts:
        typer.echo(f"  - {script}")


@app.command()
@logger.catch
def download(script_id: str):
    """
    Downloads a script from the database.
    """
    if not check_config():
        return
    from scoach.models import Script
    script: Script = safe_object_get(Script, id=script_id)
    if script is None:
        typer.echo(f"Script {script_id} not found.")
        return
    typer.echo(f"Script {script_id} is available! Downloading...")
    minio_client = get_minio_client()
    download_from_minio(minio_client, script.path,
                        f"{script.id}.py")


@app.command()
@logger.catch
def delete(script_id: str):
    """
    Deletes a script from the database.
    """
    if not check_config():
        return
    from scoach.models import Script
    script: Script = safe_object_get(Script, id=script_id)
    if script is None:
        typer.echo(f"Script {script_id} not found.")
        return
    typer.echo(f"Script {script_id} will be deleted!")
    typer.echo(f"Information: {script}")
    choice = typer.prompt("Are you sure?", default=False,
                          show_choices=True, type=bool)
    if choice:
        script.delete()
        typer.echo(f"Script {script_id} deleted!")
    else:
        typer.echo("Script not deleted.")


@app.command()
@logger.catch
def describe(script_id: str):
    """
    Describes a script in the database.
    """
    if not check_config():
        return
    from scoach.models import Script, Run
    script: Script = safe_object_get(Script, id=script_id)
    if script is None:
        typer.echo(f"Script {script_id} not found.")
        return
    typer.echo(f"Script #{script_id}")
    typer.echo(f"  - Created at: {script.date_created}")
    typer.echo(f"  - Modified at: {script.date_modified}")
    typer.echo(f"  - Path: {script.path}")
    related_runs = Run.objects.filter(script=script)
    if len(related_runs) == 0:
        typer.echo("  - No runs found for this script.")
    else:
        typer.echo(f"  - {len(related_runs)} runs found for this script:")
        for run in related_runs:
            typer.echo(f"    * {run}")
    typer.echo()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scoach', 'scoach.cli', 'scoach.settings']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.2.0,<4.0.0',
 'Jinja2>=3.0.0,<4.0.0',
 'PyYAML>=5.4.0,<6.0.0',
 'dask_jobqueue>=0.7.0,<0.8.0',
 'loguru>=0.5.0,<0.6.0',
 'minio>=7.0.0,<8.0.0',
 'prefect>=0.15.0,<0.16.0',
 'psycopg2-binary>=2.9.0,<3.0.0',
 'tensorflow>=2.0.0,<3.0.0',
 'typer>=0.4.0,<0.5.0']

entry_points = \
{'console_scripts': ['scoach = scoach.cli:app']}

setup_kwargs = {
    'name': 'scoach',
    'version': '0.1.0',
    'description': 'Setup for training Tensorflow models on SLURM clusters.',
    'long_description': "# scoach\n\nA setup for training Tensorflow models on SLURM clusters\n\n## How does it work?\n\n- Inputs needed (see examples directory):\n  - A `.json` file with parameters for training\n  - A `.json` file with the model definition\n  - A `.py` file with the training code.\n- There's a CLI app for interacting with scoach\n- Run `scoach init` for setting up your configuration file, such as in `config_example.yaml`\n- On the login machine at the SLURM cluster, run `scoach start`. This will start a daemon that will then launch jobs as requested.\n- On any machine, you can do `scoach run submit` to submit jobs.\n  - This will upload the Python script to MinIO and submit the configurations to the database.\n  - The new runs are consumed by the daemon process, which then uses Jinja2 to render the training script and submit it to the cluster.\n  - The training script is then run on the cluster, using Dask workers, that will grow as needed.\n\n## To do\n\n- [ ] Add support for uploading/managing datasets\n- [ ] No Python script duplicates\n",
    'author': 'Gabriel Gazola Milan',
    'author_email': 'gabriel.gazola@poli.ufrj.br',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/gabriel-milan/scoach',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import os

# This defines the base dir for all relative imports for our project, put the file in your root folder so the
# base_dir points to the root folder
BASE_DIR = os.path.dirname(os.path.abspath(__file__))

DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'

# Since we only have one app which we use
INSTALLED_APPS = (
    'scoach',
)



from .base import *

# This is not a secret!!!
SECRET_KEY = 'thisisnotasecret'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}
